"""SQL Session product


CREATE TABLE ZSessions (
  zsession varchar(40) UNIQUE NOT NULL,
   -- time?
);

CREATE TABLE ZSession_data (
  zsession varchar(40) NOT NULL,
  zname varchar(200) NOT NULL,
  zvalue varchar(4000),
  primary key (zsession, zname)
);


"""

__version__='$Revision: 1.17 $'[11:-2]

import Globals, App.Undo, socket, regex
from Products.ZSQLMethods.SQL import SQL
from Globals import PersistentMapping
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from string import join,strip,split,lower
from App.Management import Navigation, Tabs
import Acquisition
from Acquisition import Implicit, Explicit
import OFS.SimpleItem 
from OFS.Document import Document
from urllib import quote, unquote
from ImageFile import ImageFile
import AccessControl.Role 
import string
from cPickle import loads, dumps
from Shared.DC.ZRDB.TM import TM, Surrogate

ListType=type([])

class SpecialItems: pass

DeletedItem = SpecialItems()



class SessionObj(Explicit, TM):
    """ A session object. Installed into the REQUEST object as 
	REQUEST.SESSION (like REQUEST.RESPONSE)
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def _begin(self, *args):
	"begin method. does diddly"
	#print "begin called for session obj", id(self), self.sName
        pass

    def _abort(self, *args):
	""" abort method. cleans up local storage. (removes loop to 
	    parentSession)
	"""
	#print "abort called for session obj", id(self), self.sName
	del self.cachedict
	del self.changedict
	del self.parentSession
    
    def _finish(self, *args):
	""" finish. store stuff in DB, clean up local storage (remove loop to
	    parentSession
	"""
	#print "finish called for session obj", id(self), self.sName
	if self.was_cleared:
	    self.parentSession.sqlDeleteValue(session=self.sName)
	    dict = self.cachedict
	else:
	    dict = self.changedict
	for name in dict.keys():
	    value = dict[name]
	    if value is DeletedItem:
		self.parentSession.sqlDeleteValue(session=self.sName, name=name)
	    else:
		encvalue = self.dumper(value)
		self.parentSession.sqlSetValue(session=self.sName,
				    name=name, value=encvalue)
	del self.cachedict
	del self.changedict
	del self.parentSession

    def _registerme(self):
        if not self._registered:
            try:
		try:
		    get_transaction().register(Surrogate(self), early=1)
		    self._begin()
		    self._registered=1
		except TypeError: # unpatched ZODB/Transaction.py
		    self._register()
            except: pass
	
    def __init__(self, parentSession, sessionName, mode='base64'):
	#print "sessionobj init called"
	self.sName = sessionName
	self.cachedict = {}
	self.changedict = {}
	self.parentSession = parentSession # argh loop.
	self.was_cleared = 0 # set to 1 when a 'clear' operation happens.
	self.loadall_complete = 0
	self._registerme()

    def getName(self):
	"return name of session"
	return self.sName

    def loader(self, v):
	"load - base64"
        return loads(decodestring(v))

    def dumper(self, v):
	"dump - base64"
        return encodestring(dumps(v))[:-1] # chop \n

    def precache(self):
	if not self.loadall_complete:
	    rows = self.parentSession.sqlGetValue(session=self.sName)
	    for r in rows:
		n = sqlattr(r, 'name')
		v = self.loader(sqlattr(r, 'value'))
		self.cachedict[n] = v
	    self.loadall_complete = 1

    def __getitem__(self, name):
        if self.cachedict.has_key(name):
	    return self.cachedict[name]
	else:
	    raise KeyError,name

    def __setitem__(self, name, value):
	self.cachedict[name] = value
	self.changedict[name] = value
	return None

    set = __setitem__

    def __delitem__(self, name):
	del self.cachedict[name]
	self.changedict[name] = DeletedItem

    delete = __delitem__
    
    def __len__(self):
	rows = self.cachedict.keys()
        return len(rows)

    def keys(self):
	return self.cachedict.keys()

    def values(self):
	return self.cachedict.values()

    def items(self):
	return self.cachedict.items()

    def has_key(self, name):
	return self.cachedict.has_key(name)

    def __str__(self):
	l = ["Session Variables:\n"]
	for n,v in self.items():
	    l.append("%s: %s"%(n,str(v)))
	return join(l, '\n')

    def clear(self):
	self.cachedict = {}
	self.changedict = {}
	self.was_cleared = 1 # we have to do a complete reload.

    def copy(self):
        raise RuntimeError, "copy() unimplemented for sessions"
    
    def update(self, dict):
        for k, v in dict.items():
            self[k] = v

    def get(self, key, failobj=None):
        if self.has_key(key): return self[key]
        else: return failobj


class SQLSession(OFS.SimpleItem.Item, Persistent, Implicit, 
		      AccessControl.Role.RoleManager):
    """ An SQL Session. """


    meta_type='SQL Session'
    id       ='Session'
    title    ='SQL Session'
    icon     ='misc_/SQLSession/SQLSession_icon'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_tabs' ]),
    ('Change SQLSessions',     ['manage_edit']),
    ('View', ['__call__','']),
    )

    manage_main=HTMLFile('main', globals())

    def __init__(self, conn_id, cookie_name, mode=None, path='/'):
        self.conn_id=conn_id
	self.cookie_name = cookie_name
	if not mode: mode = 'base64'
	self.mode = mode
	self.path = path
        
        #self.docLogin =Document(_docLogin, __name__='docLogin')
        #self.docLogout=Document(_docLogout, __name__='docLogout')
        
        self.sqlSessionCreate=SQL('sqlSessionCreate', '', conn_id,
                              'session',
                              _sqlSessionCreate)
        self.sqlSessionLookup=SQL('sqlSessionLookup', '', conn_id,
                              'session',
                              _sqlSessionLookup)
        self.sqlGetValue=SQL('sqlGetValue', '', conn_id,
                              'session name=""',
                              _sqlGetValue)
        self.sqlSetValue=SQL('sqlSetValue', '', conn_id,
                             'session name value',
                              _sqlSetValue)
        self.sqlDeleteValue=SQL('sqlDeleteValue', '', conn_id,
                              'session name=""',
                              _sqlDeleteValue)

    def enumQueries(self):
        list=[]
        type=SQL.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def enumDocuments(self):
        list=[]
        type=Document.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def manage_edit(self, connection_id, cookie_name, mode, path, REQUEST=None):
	"""Change properties"""
        if connection_id != self.conn_id:
            # Connection changed - update sql methods!
            self.conn_id=connection_id
            for ob in self.enumQueries():
                title=ob.title
                args=ob.arguments_src
                src=ob.src
                ob.manage_edit(title, self.conn_id, args, src)
	self.cookie_name = cookie_name
	self.mode = mode
	self.path = path
	self.check_and_update()
        if REQUEST:
            return MessageDialog(
                   title  ='SQLSesssion Changed',
                   message='SQLSesssion properties have been updated',
                   action =REQUEST['URL1']+'/manage_main',
                   target ='manage_main')

    def check_and_update(self):
	"upgrade in-place sessions"
	try:
	    d = self.sqlDeleteValue
        except:
            self.sqlDeleteValue=SQL('sqlDeleteValue', '', self.conn_id,
                               'session name',
                               _sqlDeleteValue)

    def __call__(self, REQUEST=None, session_id=None, noCookie=None, validSession=None):
	""" magic bits here. """
	#print "called"
	if REQUEST is None:
	    try:
		REQUEST=self.REQUEST
	    except:
		raise "Session Error", "can't find REQUEST, even in self!"
	response = REQUEST['RESPONSE']

	if ( REQUEST.has_key('SESSION') ):
	    if not ( REQUEST.has_key('_force_new_session') and 
			    REQUEST['_force_new_session'] ):
		return

	if session_id is not None:
	    if not REQUEST.cookies.has_key(self.cookie_name):
		sessionName = self.newSession(REQUEST, session_id, noCookie)
	    sessionName = REQUEST.cookies[self.cookie_name]
	    if sessionName != session_id:
		sessionName = self.newSession(REQUEST, session_id, noCookie)
	    r = self.sqlSessionLookup(session=sessionName)
	    if len(r) != 1:
		if validSession: # validSession means don't autocreate a session
		    raise "Session Error", "Invalid session specified"
		else:
		    sessionName = self.newSession(REQUEST, noCookie=noCookie)
	    else:
		pass # session is fine as-is.
	else:
	    if not REQUEST.cookies.has_key(self.cookie_name):
		sessionName = self.newSession(REQUEST, noCookie=noCookie)
	    else:
		if ( REQUEST.has_key('_force_new_session') and 
					    REQUEST['_force_new_session'] ):
		    sessionName = self.newSession(REQUEST, noCookie=noCookie)
		    REQUEST.set('_force_new_session',0)
		else:
		    sessionName = REQUEST.cookies[self.cookie_name]
		    r = self.sqlSessionLookup(session=sessionName)
		    if len(r) != 1:
			if validSession: 
			    raise "Session Error", "Invalid session specified"
			else:
			    sessionName = self.newSession(REQUEST, noCookie=noCookie)
		    else:
			pass # session is fine as-is.

	self.buildSessionObj(REQUEST, sessionName)
	return '<!-- session is "%s" -->'%sessionName

    def buildSessionObj(self, REQUEST, sessionName):
	s = SessionObj(self, sessionName, self.mode)
	REQUEST['SESSION'] = s.__of__(self)
	REQUEST['SESSION'].precache()

    def newSession(self, REQUEST, sessionName=None, noCookie=None):
	" create a new session. set the cookie, all that stuff. "
	response = REQUEST['RESPONSE']
	if sessionName is None:
	    sessionName = self.makeSessionName(REQUEST)
	    r = self.sqlSessionCreate(session=sessionName)
	# on some browsers, doing an expire and a set in the same
	# request is a no-op. :-(
	#response.expireCookie(self.cookie_name, path=self.path)
	if not noCookie:
	    response.setCookie(self.cookie_name, sessionName, path=self.path)
	REQUEST.cookies[self.cookie_name] = sessionName
	return sessionName

    def makeSessionName(self, REQUEST):
	""" create a new unique string to name the session. feel free to
	    fiddle with this. """
	import md5, time, base64
	m = md5.new()
	m.update('happy happy happy boing plonk splut')
	m.update(str(time.time()))
	m.update(str(REQUEST))
	return string.replace(base64.encodestring(m.digest())[:-3], '/', '$')



manage_addSQLSessionForm=HTMLFile('addSQLSession', globals())

def manage_addSQLSession(self, connection_id, cookie_name, REQUEST=None):
    """ """
    if hasattr(self.aq_base, 'Session'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a Session',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    ob=SQLSession(connection_id, cookie_name)
    self._setObject('Session', ob)
    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)





# A getattr-like helper that hopefully can compensate for the
# various oddities in data returned by different stupidbases.


import Missing
mt=type(Missing.Value)

def typeconv(val):
    if type(val)==mt:
        return ''
    return val

def sqlattr(ob, attr):
    name=attr
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=string.upper(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=string.lower(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    raise NameError, name

def absattr(attr):
    if callable(attr): return attr()
    return attr

def reqattr(request, attr):
    try:    return request[attr]
    except: return None

_sqlSessionCreate = """<dtml-comment>

This SQL creates a new session.

</dtml-comment>
INSERT INTO ZSessions 
( zsession )
values
( <dtml-sqlvar session type="string"> )

"""

_sqlSessionLookup = """<dtml-comment>
This query looks up a particular session. It 
should return one or more rows if the session 
is active.
</dtml-comment>

SELECT zsession as "session" 
FROM ZSessions
WHERE 
<dtml-sqltest session column=zsession type=string>

"""

_sqlGetValue = """<dtml-comment>
This query looks up some or all data belonging 
to a session.
</dtml-comment>

SELECT zsession as "session",
	zname as "name",
	zvalue as "value" 
FROM ZSession_data
WHERE 
<dtml-sqltest session column=zsession type=string>
<dtml-if name>
AND <dtml-sqltest name column=zname type=string>
</dtml-if name>

"""

_sqlSetValue = """<dtml-comment>
This query sets or updates a data item belonging 
to a session.  Updates are bad, so we delete/insert.
</dtml-comment>

DELETE FROM ZSession_data
WHERE 
<dtml-sqltest session column=zsession type=string>
AND 
<dtml-sqltest name column=zname type=string>

<dtml-var sql_delimiter>

INSERT INTO ZSession_data
( zsession, zname, zvalue ) 
VALUES
( <dtml-sqlvar session type="string">,
  <dtml-sqlvar name type="string">,
  <dtml-sqlvar value type="string"> )

"""

_sqlDeleteValue = """<dtml-comment>
This query deletes a data item for a session.
</dtml-comment>

DELETE FROM ZSession_data
WHERE 
<dtml-sqltest session column=zsession type=string>
<dtml-if name>
AND 
<dtml-sqltest name column=zname type=string>
</dtml-if>

"""


import binascii, base64, string

def encodestring(bin):
    b2a = binascii.b2a_base64
    MAXLINESIZE = 76 # Excluding the CRLF
    MAXBINSIZE = (MAXLINESIZE/4)*3
    if len(bin) > MAXBINSIZE:
	out = []
	for o in range(0, len(bin), MAXBINSIZE):
	    out.append(b2a(bin[o:o+MAXBINSIZE]))
	asc = string.join(out,'')
    else:
	asc = b2a(bin)
    return asc

def decodestring(asc):
    a2b = binascii.a2b_base64
    bin = a2b(asc)
    return bin
